/**
 ** Ba**** Camera image to PPM convertor.
 **/
/*
 * PPM looks like this:
 *
 * P3       <---- magic
 * 327 360  <---- ncols & nrows
 * 255      <---- maxval
 * 255 0 255  255 0 255  255 0 255  255 0 255  255 0 255  255 0 255
 * ....
 * 255 0 255  255 0 255  255 0 255
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#define BCAM_NCOLS   164
#define BCAM_NROWS   126
#define BCAM_BLIND     2	/* 2 blank lines */
#define BCAM_SIZE	(BCAM_NCOLS*BCAM_NROWS)

#define NAME   "bcam2ppm"

struct pixel {
	unsigned char r, g, b;
	unsigned char valid;
};

void convert_small(unsigned char *, FILE *);

int main()
{
	unsigned char *buf;
	FILE *ifile = stdin;
	FILE *ofile = stdout;

	if ((buf = malloc(BCAM_SIZE)) == NULL) {
		fprintf(stderr, NAME ": no core\n");
		exit(1);
	}

	if (fread(buf, 1, BCAM_SIZE, ifile) != BCAM_SIZE) {
		fprintf(stderr, NAME ": short read\n");
		exit(1);
	}

	convert_small(buf, ofile);

	free(buf);
	exit(0);
	return 0;
}

/*
 * convert_small() takes three elements of the Bayer pattern that
 * the camera delivers and combines them into a single pixel,
 * producing a picture that is twice smaller (80 x 60).
 */
void convert_small(unsigned char *buf, FILE *ofile) {
	struct pixel *frame, *p;
	struct pixel max;
	int row, col;
	int xpos;
	char *cp;
	int maxmod;

	/*
	 * XXX We allocate a tad more due to two blind lines....
	 */
	if ((frame = malloc(sizeof(struct pixel) * (BCAM_SIZE/4))) == NULL) {
		fprintf(stderr, NAME ": no core\n");
		exit(1);
	}
	memset((void *)frame, 0, sizeof(struct pixel) * (BCAM_SIZE/4));

	max.r = max.g = max.b = 0;
	max.valid = 1;

	/*
	 * Collect, interleaving odd and even lines.
	 */
	buf += BCAM_NCOLS*BCAM_BLIND;	/* skip blind lines */
	p = frame;
	for (row = 0; row < (BCAM_NROWS-BCAM_BLIND)/2; row++) {
		for (col = 0; col < BCAM_NCOLS/2; col++) {
			/* p = &frame[row * (BCAM_NCOLS/2) + col]; */

			cp = &buf[2*row * BCAM_NCOLS + col];
			p->r = cp[0];
			p->g = cp[BCAM_NCOLS/2];   /* Also [BCAM_NCOLS] */
			p->b = cp[BCAM_NCOLS + BCAM_NCOLS/2];
			p->valid = 1;

			if (p->r > max.r) max.r = p->r;
			if (p->g > max.g) max.g = p->g;
			if (p->b > max.b) max.b = p->b;

			++p;
		}
	}

/* P3 */ fprintf(stderr, "max r %d g %d b %d\n", max.r, max.g, max.b);

	fprintf(ofile, "P3\n");
	fprintf(ofile, "%d %d\n", BCAM_NCOLS/2, (BCAM_NROWS-BCAM_BLIND)/2);

	maxmod = max.r;
	if (max.g > maxmod) maxmod = max.g;
	if (max.b > maxmod) maxmod = max.b;
	fprintf(ofile, "%d\n", maxmod);
	/* fprintf(ofile, "255\n"); */

	xpos = 0;

	p = frame;
	for (row = 0; row < (BCAM_NROWS-BCAM_BLIND)/2; row++) {
		for (col = 0; col < BCAM_NCOLS/2; col++) {

			if (!p->valid) {
				fprintf(stderr, "IE: invalid pixel at "
				    "row %d column %d\n", row, col);
				exit(1);
			}

			if (xpos) {
				printf(" ");
			}
			fprintf(ofile, "%d %d %d", p->r, p->g, p->b);
			if (++xpos >= 6) {
				xpos = 0;
				fprintf(ofile, "\n");
			}

			++p;
		}
	}
	if (xpos) {
		fprintf(ofile, "\n");
	}

	free(frame);
}
