/*
 * miniannex
 * This is a minimal daemon that allows to access remote serial ports.
 * It is modelled after Xylogics Annex which listens on port 5000+n.
 *
 * (c) Copyright 2000 Free Software Foundation
 * Author: Pete Zaitcev
 *
 * This software comes with ABSOLUTELY NO WARRANTY. Use it on your own risk.
 *
 * This program is a TCP server but it does not demonise itself
 * properly. It is just too simple for that hassle.
 * Use a shell with job control to start miniannex.
 */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>
#include <netinet/in.h>

void Usage(void);
#if 0
int read_all(int fd, char *data, int size);
#endif

#define NAME  "miniannex"
#define SBSIZE     1200		/* Let it fit into an Ethernet frame */

#define PROTO_TCP    6		/* == getprotobyname("tcp") */
#define DFL_PORT   5000		/* Annex-like TCP port */

/*
 * Open tty and set serial parameters.
 */
int opentty(char *name) {
	int dfd;
	struct termios tb;

	if ((dfd = open(name, O_RDWR|O_NONBLOCK)) < 0) {
		fprintf(stderr, NAME ": Cannot open %s: %s\n", name,
		    strerror(errno));
		exit(1);
	}

	(void) ioctl(dfd, TCFLSH, TCIOFLUSH);

	memset(&tb, 0, sizeof(tb));
	tb.c_iflag = 0;
	tb.c_oflag = 0;
	tb.c_cflag = CS8 | CREAD | CLOCAL | B57600;	/* HUPCL? */
	tb.c_lflag = 0;
	tb.c_cc[VMIN] = 1;
	tb.c_cc[VTIME] = 0;
	if (ioctl(dfd, TCSETS, &tb) != 0) {
		fprintf(stderr, NAME ": Cannot set tty %s: %s\n", name,
		    strerror(errno));
		exit(1);
	}

	/* make the serial port blocking again */
	(void) fcntl(dfd, F_SETFL, fcntl(dfd, F_GETFL, 0) & ~O_NONBLOCK);

	return dfd;
}

/*
 * socket, bind, listen, but do not accept yet.
 */
int opennet(int port) {
	int afd;
	struct sockaddr_in loca;

	if ((afd = socket(AF_INET, SOCK_STREAM, PROTO_TCP)) < 0) {
		fprintf(stderr, NAME ": socket: %s\n", strerror(errno));
		exit(1);
	}

	memset(&loca, 0, sizeof(loca));
	loca.sin_family = AF_INET;
	loca.sin_port = htons(port);
	loca.sin_addr.s_addr = INADDR_ANY;	/* htonl */

	if (bind(afd, (struct sockaddr *)&loca, sizeof(loca)) < 0) {
		fprintf(stderr, NAME ": bind: %s\n", strerror(errno));
		exit(1);
	}

	/* We do not allow many clients, so need not backlog. */
	listen(afd, 1);

	return afd;
}

/*
 * Do the accept and abort if there is a trouble.
 */
int ac_one(int afd) {
	struct sockaddr addr;
	int sfd;
	int asize;

	asize = sizeof(addr);
	if ((sfd = accept(afd, &addr, &asize)) < 0) {
		fprintf(stderr, NAME ": accept: %s\n", strerror(errno));
		exit(1);
	}

	return sfd;
}

/*
 * Basically, a call to atoi().
 */
int getport(char *pstr) {
	char *s;
	int port;

	for (s = pstr; *s != 0; s++) {
		if (!isdigit(*s)) goto bad;
	}
	port = atoi(pstr);
	if (port <= 0 || port >= 32767) goto bad;
	return port;

bad:
	fprintf(stderr, NAME ": bad network port \"%s\"\n", pstr);
	exit(1);
	return -1;
}

/*
 */
void net_to_tty(int sfd, int dfd, int kidpid) {
	static char buf[SBSIZE];
	int rc;

	for (;;) {
		rc = read(sfd, buf, SBSIZE);
		if (rc < 0) {
			kill(kidpid, SIGTERM);
			fprintf(stderr, NAME ": net read error: %s\n",
			    strerror(errno));
			exit(1);
		}
		if (rc == 0) {
			return;
		}

		if (write(dfd, buf, rc) < 0) {
			kill(kidpid, SIGTERM);
			fprintf(stderr, NAME ": port write error: %s\n",
			    strerror(errno));
			exit(1);
		}
	}
}

/*
 */
void tty_to_net(int sfd, int dfd) {
	static char buf[SBSIZE];
	int rc;

	for (;;) {
		rc = read(dfd, buf, SBSIZE);
		if (rc < 0) {
			fprintf(stderr, NAME ": port read error: %s\n",
			    strerror(errno));
			exit(1);
		}
		if (rc == 0) {	/* Should not happen, but... */
			fprintf(stderr, NAME ": port EOF\n");
			return;
		}

		if (write(sfd, buf, rc) < 0) {	/* ESPIPE usually */
			fprintf(stderr, NAME ": net write error: %s\n",
			    strerror(errno));
			exit(1);
		}
	}
}

/*
 * Well, main, duh.
 */
int main(int argc, char *argv[]) {
	int dfd, sfd, afd;
	int port = 0;
	int wstat;
	int rc;

	if (argc == 2) {
		port = DFL_PORT;
	} else if (argc == 3) {
		port = getport(argv[2]);
	} else {
		Usage();
	}

	/*
	 * We considered opening upon an accept, but this makes it
	 * impossible to process errors in a tty open.
	 */
	dfd = opentty(argv[1]);

	afd = opennet(port);

	for (;;) {
		sfd = ac_one(afd);
		(void) ioctl(dfd, TCFLSH, TCIOFLUSH);	/* do it before fork */

		if ((rc = fork()) < 0) {
			exit(2);	/* Die, nothing we can do. */
		}
		if (rc == 0) {
			tty_to_net(sfd, dfd);
			exit(0);
		}
		net_to_tty(sfd, dfd, rc);
		kill(rc, SIGTERM);	/* in case the child is idle. */
		wait(&wstat);

		close(sfd);
	}
	return 0;
}

/*
 */
void Usage(void) {
	fprintf(stderr, "Usage: miniannex /dev/tty<port> [net-port]\n");
	exit(1);
}

#if 0
/*
 * Read bytes until the requested size is satisfied.
 * Normally this is done when reading from a network socket,
 * but we must to it while reading our serial connection too
 * because we set VMIN=1. If we set it differently, some broken
 * setups stop working (reading off a terminal server, etc.).
 */
int read_all(int fd, char *data, int size) {
	int togo;
	int rc;

	togo = size;
	while (togo > 0) {
		if ((rc = read(fd, data, togo)) < 0) return -1;
		if (rc == 0) {
			/* P3 */ fprintf(stderr, "EOF reading serial\n");
			return size - togo;
		}
		togo -= rc;
		data += rc;
	}
	return size;
}
#endif
